/*
 * DVB USB Linux driver for Afatech AF9015 DVB-T USB2.0 receiver
 *
 * Copyright (C) 2007 Antti Palosaari <crope@iki.fi>
 *
 * Thanks to Afatech who kindly provided information.
 *
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#ifndef _DVB_USB_AF9015_H_
#define _DVB_USB_AF9015_H_

#define DVB_USB_LOG_PREFIX "af9015"
#include "dvb-usb.h"

extern int dvb_usb_af9015_debug;
#define deb_info(args...) dprintk(dvb_usb_af9015_debug, 0x01, args)
#define deb_xfer(args...) dprintk(dvb_usb_af9015_debug, 0x02, args)
#define deb_rc(args...)   dprintk(dvb_usb_af9015_debug, 0x04, args)
#define deb_reg(args...)  dprintk(dvb_usb_af9015_debug, 0x08, args)
#define deb_i2c(args...)  dprintk(dvb_usb_af9015_debug, 0x10, args)
#define deb_fw(args...)   dprintk(dvb_usb_af9015_debug, 0x20, args)

#define AF9015_I2C_EEPROM  0xa0
#define AF9015_I2C_DEMOD   0x38
#define AF9015_USB_TIMEOUT 1000

/* EEPROM locations */
#define AF9015_EEPROM_IR_MODE        0x18
#define AF9015_EEPROM_IR_REMOTE_TYPE 0x34

#define AF9015_EEPROM_SAW_BW1        0x35
#define AF9015_EEPROM_XTAL_TYPE1     0x36
#define AF9015_EEPROM_SPEC_INV1      0x37
#define AF9015_EEPROM_IF1L           0x38
#define AF9015_EEPROM_IF1H           0x39
#define AF9015_EEPROM_MT2060_IF1L    0x3a
#define AF9015_EEPROM_MT2060_IF1H    0x3b
#define AF9015_EEPROM_TUNER_ID1      0x3c

#define AF9015_EEPROM_SAW_BW2        0x45
#define AF9015_EEPROM_XTAL_TYPE2     0x46
#define AF9015_EEPROM_SPEC_INV2      0x47
#define AF9015_EEPROM_IF2L           0x48
#define AF9015_EEPROM_IF2H           0x49
#define AF9015_EEPROM_MT2060_IF2L    0x4a
#define AF9015_EEPROM_MT2060_IF2H    0x4b
#define AF9015_EEPROM_TUNER_ID2      0x4c

#define AF9015_GPIO_ON (1 << 0)
#define AF9015_GPIO_EN (1 << 1)
#define AF9015_GPIO_O  (1 << 2)
#define AF9015_GPIO_I  (1 << 3)

#define AF9015_GPIO_TUNER_ON  (AF9015_GPIO_ON|AF9015_GPIO_EN)
#define AF9015_GPIO_TUNER_OFF (AF9015_GPIO_ON|AF9015_GPIO_EN|AF9015_GPIO_O)

struct req_t {
	u8  cmd;       /* [0] */
	/*  seq */     /* [1] */
	u8  i2c_addr;  /* [2] */
	u16 addr;      /* [3|4] */
	u8  data0;     /* [5] */
	u8  data1;     /* [6] */
	u8  len;       /* [7] */
	u8  *data;
};

enum cmd {
	GET_CONFIG           = 0x10,
	DOWNLOAD_FIRMWARE    = 0x11,
	BOOT                 = 0x13,
	READ_MEMORY          = 0x20,
	WRITE_MEMORY         = 0x21,
	RECONNECT_USB        = 0x5a,
	WRITE_VIRTUAL_MEMORY = 0x26,
	READ_WRITE_I2C       = 0x22,
	GET_IR_CODE          = 0x27,
	READ_I2C,
	WRITE_I2C,
};

struct af9015_state {
	u8 gpio0;
	u8 gpio1;
	u8 gpio2;
	u8 gpio3;
	u16 mt2060_if1;
};

/* A-Link DTU(m) */
static u8 ir_table[] = {
	0x08, 0xf7, 0x12, 0xed, 0x45, 0x05, 0x00,
	0x08, 0xf7, 0x1a, 0xe5, 0x41, 0x00, 0x00,
	0x08, 0xf7, 0x01, 0xfe, 0x1e, 0x00, 0x00,
	0x08, 0xf7, 0x1c, 0xe3, 0x21, 0x00, 0x00,
	0x08, 0xf7, 0x03, 0xfc, 0x24, 0x00, 0x00,
	0x08, 0xf7, 0x05, 0xfa, 0x28, 0x04, 0x00,
	0x08, 0xf7, 0x00, 0xff, 0x43, 0x00, 0x00,
	0x08, 0xf7, 0x16, 0xe9, 0x42, 0x00, 0x00,
	0x08, 0xf7, 0x0f, 0xf0, 0x1f, 0x00, 0x00,
	0x08, 0xf7, 0x0d, 0xf2, 0x22, 0x00, 0x00,
	0x08, 0xf7, 0x1b, 0xe4, 0x25, 0x00, 0x00,
	0x08, 0xf7, 0x06, 0xf9, 0x27, 0x00, 0x00,
	0x08, 0xf7, 0x14, 0xeb, 0x2e, 0x00, 0x00,
	0x08, 0xf7, 0x1d, 0xe2, 0x2d, 0x00, 0x00,
	0x08, 0xf7, 0x02, 0xfd, 0x20, 0x00, 0x00,
	0x08, 0xf7, 0x18, 0xe7, 0x23, 0x00, 0x00,
	0x08, 0xf7, 0x04, 0xfb, 0x26, 0x00, 0x00,
	0x08, 0xf7, 0x07, 0xf8, 0x44, 0x00, 0x00 };

#endif
