/*
	Maxlinear MXL5003/5 Multi Mode tuner driver

	Copyright (C) Manu Abraham <abraham.manu@gmail.com>
	Copyright (C) Matthias Schwarzott <zzam@gentoo.org>

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <linux/module.h>
#include <linux/moduleparam.h>
#include <linux/delay.h>
#include <linux/dvb/frontend.h>
#include <linux/i2c.h>

#include "dvb_frontend.h"

#include "mxl500x.h"
#include "mxl500x_reg.h"

struct mxl500x_reg {
	u8 reg;
	u8 data;
};

struct mxl500x_state {
	struct i2c_adapter	*i2c;
	struct dvb_frontend	*frontend;

	struct mxl500x_reg	reg_field[128];

	const struct mxl500x_config *config;
};


// Register init
/* These are the default register values (in memory) */
static struct mxl500x_reg reg_init[] = {

	{   9, 0xf2 },
	{  11, 0x19 },
	{  12, 0x60 },
	{  13, 0x00 },
	{  14, 0x00 },
	{  15, 0xc0 },
	{  16, 0x00 },
	{  17, 0x00 },
	{  18, 0x00 },
	{  19, 0x34 },
	{  21, 0x00 },
	{  22, 0x6b },
	{  23, 0x35 },
	{  24, 0x70 },
	{  25, 0x36 },
	{  26, 0x82 },
	{  31, 0x00 },
	{  32, 0x40 },
	{  33, 0x53 },
	{  34, 0x81 },
	{  35, 0xc9 },
	{  36, 0x01 },
	{  37, 0x00 },
	{  41, 0x00 },
	{  42, 0xf8 },
	{  43, 0x43 },
	{  44, 0x20 },
	{  45, 0x80 },
	{  46, 0x88 },
	{  47, 0x86 },
	{  48, 0x00 },
	{  49, 0x00 },
	{  53, 0x94 },
	{  54, 0xfa },
	{  55, 0x92 },
	{  56, 0x80 },
	{  57, 0x41 },
	{  58, 0xdb },
	{  59, 0x00 },
	{  60, 0x00 },
	{  61, 0x00 },
	{  62, 0x00 },
	{  65, 0xf8 },
	{  66, 0xe4 },
	{  67, 0x90 },
	{  68, 0xc0 },
	{  69, 0x01 },
	{  70, 0x50 },
	{  71, 0x06 },
	{  72, 0x00 },
	{  73, 0x20 },
	{  76, 0xbb },
	{  77, 0x13 },
	{  81, 0x04 },
	{  82, 0x75 },
	{  83, 0x00 },
	{  84, 0x00 },
	{  85, 0x00 },
	{  91, 0x70 },
	{  92, 0x00 },
	{  93, 0x00 },
	{  94, 0x00 },
	{  95, 0x0c },
	{  96, 0x00 },
	{  97, 0x00 },
	{  98, 0xe2 },
	{  99, 0x00 },
	{ 100, 0x00 },
	{ 101, 0x12 },
	{ 102, 0x80 },
	{ 103, 0x32 },
	{ 104, 0xb4 },
	{ 105, 0x60 },
	{ 106, 0x83 },
	{ 107, 0x84 },
	{ 108, 0x9c },
	{ 109, 0x02 },
	{ 110, 0x81 },
	{ 111, 0xc0 },
	{ 112, 0x10 },
	{ 131, 0x8a },
	{ 132, 0x10 },
	{ 133, 0x24 },
	{ 134, 0x00 },
	{ 135, 0x00 },
	{ 136, 0x7e },
	{ 137, 0x40 },
	{ 138, 0x38 },
	{ 146, 0xf6 },
	{ 147, 0x1a },
	{ 148, 0x62 },
	{ 149, 0x33 },
	{ 150, 0x80 },
	{ 156, 0x56 },
	{ 157, 0x17 },
	{ 158, 0xa9 },
	{ 159, 0x00 },
	{ 160, 0x00 },
	{ 161, 0x00 },
	{ 162, 0x40 },
	{ 166, 0xae },
	{ 167, 0x01 },
	{ 168, 0xf2 },
	{ 195, 0x00 },
	{ }
};


static int mxl500x_release(struct dvb_frontend *fe)
{
	kfree(fe->tuner_priv);
	fe->tuner_priv = NULL;

	return 0;
}

enum mxl_latch_t {
	MXL_NO_LATCH,
	MXL_LATCH,
};

#define MXL500x_LOWIF_TH	6280000

static int mxl500x_write(struct mxl500x_state *state, u8 reg, u8 data, enum mxl_latch_t latch)
{
	struct dvb_frontend *fe = state->frontend;
	const struct mxl500x_config *config = state->config;

	u8 buf[3] = { reg, data, 0xfe };
	struct i2c_msg msg = {
		.addr	= config->addr,
		.flags	= 0,
		.buf	= buf,
		.len	= 2,
	};
	int j;

	if (MXL_LATCH == latch)
		msg.len = 3;

	if (fe->ops.i2c_gate_ctrl) {
		printk("%s: Enable gate\n", __func__);
		if (fe->ops.i2c_gate_ctrl(fe, 1))
			goto exit;
	}
	printk("tuner access: >> [");
	for (j = 0; j < msg.len; j++)
		printk(" %02x", buf[j]);

	printk(" ]\n");
	i2c_transfer(state->i2c, &msg, 1);
	if (fe->ops.i2c_gate_ctrl) {
		printk("%s: disable gate\n", __func__);
		if (fe->ops.i2c_gate_ctrl(fe, 0))
			goto exit;
	}

	return 0;
exit:
	printk("%s: I/O Error\n", __func__);
	return -EREMOTEIO;
}

static u8 mxl500x_get_reg(struct mxl500x_state *state, u8 reg);

static int mxl500x_write_regs(struct mxl500x_state *state, u8 *regs, int count)
{
#define max_regs 9
	u8 buf[max_regs*2+1];
	int i;
	int reg_nr;
	int j;

	struct dvb_frontend *fe = state->frontend;
	const struct mxl500x_config *config = state->config;

	struct i2c_msg msg = {
		.addr	= config->addr,
		.flags	= 0,
		.buf	= buf,
		.len	= 3
	};

	if (fe->ops.i2c_gate_ctrl) {
		printk("%s: Enable gate\n", __func__);
		if (fe->ops.i2c_gate_ctrl(fe, 1))
			goto exit;
	}
	printk("%s: Registers to Write=%d\n", __func__, count);
	/* Look at the regs, copy those regs from the field map to the xfer buffer */
	reg_nr = 0;
	for (i = 0; i < count; i++) {
		buf[2 * reg_nr + 0] = regs[i];
		buf[2 * reg_nr + 1] = mxl500x_get_reg(state, regs[i]);
		reg_nr++;

		if ((reg_nr == max_regs) || (i == (count - 1))) {
			msg.len = reg_nr * 2;
			if (i == (count - 1)) {
				/* latch transaction */
				msg.buf[msg.len] = 0xfe;
				msg.len++;
			}

			printk("tuner access: >> [");
			for (j = 0; j < msg.len; j++)
				printk(" %02x", buf[j]);

			printk(" ]\n");

			i2c_transfer(state->i2c, &msg, 1);
			msleep(1);
			reg_nr = 0;
		}
	}
	if (fe->ops.i2c_gate_ctrl) {
		printk("%s: Disable gate\n", __func__);
		if (fe->ops.i2c_gate_ctrl(fe, 0))
			goto exit;
	}

	return 0;
exit:
	printk("%s: I/O Error\n", __func__);
	return -EREMOTEIO;
}

static u8 mxl500x_get_reg(struct mxl500x_state *state, u8 reg)
{
	struct mxl500x_reg *reg_field = state->reg_field;
	int i = 0;

	while (reg_field[i].reg != reg)
		i++;
	return reg_field[i].data;
}

static void mxl500x_set_reg(struct mxl500x_state *state, u8 reg, u8 data)
{
	struct mxl500x_reg *reg_field = state->reg_field;
	int i = 0;

	while (reg_field[i].reg != reg)
		i++;
	reg_field[i].data = data;

}

static int mxl500x_init(struct dvb_frontend *fe)
{

	return 0;
}

#define MXL500x_SET_MAP(__reg, __field, __data)	({				\
	u8 __tmp = mxl500x_get_reg(state, __reg);				\
										\
	(__tmp = (__tmp & (~(((1 << MXL500x_WIDTH_##__field) - 1) <<		\
	MXL500x_OFFST_##__field))) | (__data << MXL500x_OFFST_##__field));	\
										\
	mxl500x_set_reg(state, __reg, __tmp);					\
})

static void mxl500x_set_gpio(struct mxl500x_state *state, u8 gpio, u8 data)
{
	if (gpio == 1)
		MXL500x_SET_MAP(MXL500x_MISC_TUNE3, MISC_TUNE3_GPIO_1B, data ? 0:1);

	// GPIO2 is not available
	if (gpio == 3) {
		if (data == 1) {
			MXL500x_SET_MAP(MXL500x_MISC_TUNE3, MISC_TUNE3_GPIO_3, 0x0);
			MXL500x_SET_MAP(MXL500x_MISC_TUNE3, MISC_TUNE3_GPIO_3B, 0x0);
		}
		if (data == 0) {
			MXL500x_SET_MAP(MXL500x_MISC_TUNE3, MISC_TUNE3_GPIO_3, 0x1);
			MXL500x_SET_MAP(MXL500x_MISC_TUNE3, MISC_TUNE3_GPIO_3B, 0x1);
		}
		if (data == 3) { /* Tri-State */
			MXL500x_SET_MAP(MXL500x_MISC_TUNE3, MISC_TUNE3_GPIO_3, 0x0);
			MXL500x_SET_MAP(MXL500x_MISC_TUNE3, MISC_TUNE3_GPIO_3B, 0x1);
		}
	}
	if (gpio == 4) {
		if (data == 1) {
			MXL500x_SET_MAP(MXL500x_MISC_TUNE3, MISC_TUNE3_GPIO_4, 0x0);
			MXL500x_SET_MAP(MXL500x_MISC_TUNE3, MISC_TUNE3_GPIO_4B, 0x0);
		}
		if (data == 0) {
			MXL500x_SET_MAP(MXL500x_MISC_TUNE3, MISC_TUNE3_GPIO_4, 0x1);
			MXL500x_SET_MAP(MXL500x_MISC_TUNE3, MISC_TUNE3_GPIO_4B, 0x1);
		}
		if (data == 3) { /* Tri-State */
			MXL500x_SET_MAP(MXL500x_MISC_TUNE3, MISC_TUNE3_GPIO_4, 0x0);
			MXL500x_SET_MAP(MXL500x_MISC_TUNE3, MISC_TUNE3_GPIO_4B, 0x1);
		}
	}
}

/* Init Registers */
static u8 mxl500x_init_regs[] = {
	MXL500x_IFSYN1,		/*  11 */
	MXL500x_IFSYN2,		/*  12 */
	MXL500x_IFSYN3,		/*  13 */
	MXL500x_RFSYN8,		/*  22 */
	0x20,
	MXL500x_IFSYN4,		/*  43 */
	MXL500x_IFSYN5,		/*  44 */
	MXL500x_LPF1,		/*  53 */
	MXL500x_MISC1,		/*  56 */
	MXL500x_LPF2,		/*  59 */
	MXL500x_DC4,		/*  73 */
	MXL500x_AGC1,		/*  76 */
	MXL500x_RSSI5,		/*  77 */
	MXL500x_XTAL1,		/*  91 */
	MXL500x_RSSI4,		/* 134 */
	MXL500x_MISC2,		/* 135 */
	MXL500x_DEMCLK,		/* 137 */
	MXL500x_IFUP1,		/* 147 */
	MXL500x_IFUP2,		/* 156 */
	MXL500x_RSSI1,		/* 166 */
	MXL500x_RSSI2,		/* 167 */
	MXL500x_RSSI3,		/* 168 */
	25,
};

/* ZIF change Registers */
static u8 mxl500x_zif_regs[] = {
	MXL500x_RFSYN1,		/*  14 */
	MXL500x_RFSYN2,		/*  15 */
	MXL500x_RFSYN3,		/*  16 */
	MXL500x_RFSYN4,		/*  17 */
	MXL500x_RFSYN8,		/*  22 */
	MXL500x_IFSYN4,		/*  43 */
	MXL500x_DC1,		/*  68 */
	MXL500x_DC2,		/*  69 */
	MXL500x_DC3,		/*  70 */
	MXL500x_DC4,		/*  73 */
	MXL500x_DAC1,		/*  92 */
	MXL500x_DAC2,		/*  93 */
	MXL500x_IQCALSYN1,	/* 106 */
	MXL500x_IQCALSYN2,	/* 107 */
	MXL500x_IQCALSYN3,	/* 108 */
	MXL500x_IQCALSYN4,	/* 109 */
	MXL500x_RFSYN5,		/* 110 */
	MXL500x_RFSYN6,		/* 111 */
	MXL500x_RFSYN7,		/* 112 */
	MXL500x_MISC_TUNE1,	/* 136 */
	MXL500x_MISC_TUNE2,	/* 138 */
	MXL500x_MISC_TUNE3,	/* 149 */
	0x4d,
	0xa6,
	0xa7,
	0xa8,
	MXL500x_PWRUD1,		/*   9 */
};

/* GO Registers */
static u8 mxl500x_go_regs[] = {
	MXL500x_IFSYN4,		/*  43 */
	MXL500x_MISC_TUNE1,	/* 136 */
	MXL500x_PWRUD1,		/*   9 */
};

/*
 * Off Chip Tracking Filter Control
 * Turn OFF Tracking filter, turn OFF all filter banks
 */
static void mxl500x_octfilter_off(struct mxl500x_state *state,
				  struct dvb_frontend_parameters *p)
{
	MXL500x_SET_MAP(MXL500x_DAC2, DAC2_DAC_A_ENABLE, 0x0);
	MXL500x_SET_MAP(MXL500x_DAC2, DAC2_DAC_B_ENABLE, 0x0);

	mxl500x_set_gpio(state, 3, 1); /* Bank 1 OFF */
	mxl500x_set_gpio(state, 1, 1); /* Bank 2 OFF */
	mxl500x_set_gpio(state, 4, 1); /* Bank 3 OFF */
}

/*
 * Off Chip Tracking Filter Control
 * Tracking filter C
 */
static void mxl500x_octfilter_c(struct mxl500x_state *state,
				struct dvb_frontend_parameters *p)
{
	MXL500x_SET_MAP(MXL500x_DAC2, DAC2_DAC_B_ENABLE, 0x1);
	MXL500x_SET_MAP(MXL500x_DAC1, DAC1_DAC_DIN_A, 0);

	if (p->frequency >= 43000000 && p->frequency < 150000000) {
		MXL500x_SET_MAP(MXL500x_DAC2, DAC2_DAC_A_ENABLE, 0x0); /* Bank 4 OFF */
		MXL500x_SET_MAP(MXL500x_DAC2, DAC2_DAC_DIN_B, 0x0);
		mxl500x_set_gpio(state, 3, 0); /* Bank 1 ON  */
		mxl500x_set_gpio(state, 1, 1); /* Bank 2 OFF */
		mxl500x_set_gpio(state, 4, 1); /* Bank 3 OFF */
	}
	if (p->frequency >= 150000000 && p->frequency < 280000000) {
		MXL500x_SET_MAP(MXL500x_DAC2, DAC2_DAC_A_ENABLE, 0x0); /* Bank 4 OFF */
		MXL500x_SET_MAP(MXL500x_DAC2, DAC2_DAC_DIN_B, 0x0);
		mxl500x_set_gpio(state, 3, 1); /* Bank 1 OFF */
		mxl500x_set_gpio(state, 1, 0); /* Bank 2 ON  */
		mxl500x_set_gpio(state, 4, 1); /* Bank 3 OFF */
	}
	if (p->frequency >= 280000000 && p->frequency < 360000000) {
		MXL500x_SET_MAP(MXL500x_DAC2, DAC2_DAC_A_ENABLE, 0x0); /* Bank 4 OFF */
		MXL500x_SET_MAP(MXL500x_DAC2, DAC2_DAC_DIN_B, 0x0);
		mxl500x_set_gpio(state, 3, 1); /* Bank 1 OFF */
		mxl500x_set_gpio(state, 1, 0); /* Bank 2 ON  */
		mxl500x_set_gpio(state, 4, 0); /* Bank 3 ON  */
	}
	if (p->frequency >= 360000000 && p->frequency < 560000000) {
		MXL500x_SET_MAP(MXL500x_DAC2, DAC2_DAC_A_ENABLE, 0x0); /* Bank 4 OFF */
		MXL500x_SET_MAP(MXL500x_DAC2, DAC2_DAC_DIN_B, 0x0);
		mxl500x_set_gpio(state, 3, 1); /* Bank 1 OFF */
		mxl500x_set_gpio(state, 1, 1); /* Bank 2 OFF */
		mxl500x_set_gpio(state, 4, 0); /* Bank 3 ON  */
	}
	if (p->frequency >= 560000000 && p->frequency < 580000000) {
		MXL500x_SET_MAP(MXL500x_DAC2, DAC2_DAC_A_ENABLE, 0x1); /* Bank 4 ON  */
		MXL500x_SET_MAP(MXL500x_DAC2, DAC2_DAC_DIN_B, 29);
		mxl500x_set_gpio(state, 3, 1); /* Bank 1 OFF */
		mxl500x_set_gpio(state, 1, 1); /* Bank 2 OFF */
		mxl500x_set_gpio(state, 4, 0); /* Bank 3 ON  */
	}
	if (p->frequency >= 580000000 && p->frequency < 630000000) {
		MXL500x_SET_MAP(MXL500x_DAC2, DAC2_DAC_A_ENABLE, 0x1); /* Bank 4 ON  */
		MXL500x_SET_MAP(MXL500x_DAC2, DAC2_DAC_DIN_B, 0);
		mxl500x_set_gpio(state, 3, 1); /* Bank 1 OFF */
		mxl500x_set_gpio(state, 1, 1); /* Bank 2 OFF */
		mxl500x_set_gpio(state, 4, 0); /* Bank 3 ON  */
	}
	if (p->frequency >= 630000000 && p->frequency < 700000000) {
		MXL500x_SET_MAP(MXL500x_DAC2, DAC2_DAC_A_ENABLE, 0x1); /* Bank 4 ON  */
		MXL500x_SET_MAP(MXL500x_DAC2, DAC2_DAC_DIN_B, 16);
		mxl500x_set_gpio(state, 3, 1); /* Bank 1 OFF */
		mxl500x_set_gpio(state, 1, 1); /* Bank 2 OFF */
		mxl500x_set_gpio(state, 4, 1); /* Bank 3 OFF */
	}
	if (p->frequency >= 700000000 && p->frequency < 760000000) {
		MXL500x_SET_MAP(MXL500x_DAC2, DAC2_DAC_A_ENABLE, 0x1); /* Bank 4 ON */
		MXL500x_SET_MAP(MXL500x_DAC2, DAC2_DAC_DIN_B, 7);
		mxl500x_set_gpio(state, 3, 1); /* Bank 1 OFF */
		mxl500x_set_gpio(state, 1, 1); /* Bank 2 OFF */
		mxl500x_set_gpio(state, 4, 1); /* Bank 3 OFF */
	}
	if (p->frequency >= 760000000 && p->frequency <= 900000000) {
		MXL500x_SET_MAP(MXL500x_DAC2, DAC2_DAC_A_ENABLE, 0x1); /* Bank 4 ON */
		MXL500x_SET_MAP(MXL500x_DAC2, DAC2_DAC_DIN_B, 0);
		mxl500x_set_gpio(state, 3, 1); /* Bank 1 OFF */
		mxl500x_set_gpio(state, 1, 1); /* Bank 2 OFF */
		mxl500x_set_gpio(state, 4, 1); /* Bank 3 OFF */
	}
}

/*
 * Off Chip Tracking Filter Control
 * Tracking filter C-H
 */
static void mxl500x_octfilter_ch(struct mxl500x_state *state,
				 struct dvb_frontend_parameters *p)
{
	MXL500x_SET_MAP(MXL500x_DAC1, DAC1_DAC_DIN_A, 0);

	if (p->frequency >= 43000000 && p->frequency < 150000000) {
		MXL500x_SET_MAP(MXL500x_DAC2, DAC2_DAC_A_ENABLE, 0x0); /* Bank 4 OFF */
		mxl500x_set_gpio(state, 4, 0); /* Bank 1 ON  */
		mxl500x_set_gpio(state, 3, 1); /* Bank 2 OFF */
		mxl500x_set_gpio(state, 1, 1); /* Bank 3 OFF */
	}
	if (p->frequency >= 150000000 && p->frequency < 280000000) {
		MXL500x_SET_MAP(MXL500x_DAC2, DAC2_DAC_A_ENABLE, 0x0); /* Bank 4 OFF */
		mxl500x_set_gpio(state, 4, 1); /* Bank 1 ON  */
		mxl500x_set_gpio(state, 3, 0); /* Bank 2 ON */
		mxl500x_set_gpio(state, 1, 1); /* Bank 3 OFF */
	}
	if (p->frequency >= 280000000 && p->frequency < 360000000) {
		MXL500x_SET_MAP(MXL500x_DAC2, DAC2_DAC_A_ENABLE, 0x0); /* Bank 4 OFF */
		mxl500x_set_gpio(state, 4, 1); /* Bank 1 OFF */
		mxl500x_set_gpio(state, 3, 0); /* Bank 2 ON  */
		mxl500x_set_gpio(state, 1, 0); /* Bank 3 ON  */
	}
	if (p->frequency >= 360000000 && p->frequency < 560000000) {
		MXL500x_SET_MAP(MXL500x_DAC2, DAC2_DAC_A_ENABLE, 0x0); /* Bank 4 OFF */
		mxl500x_set_gpio(state, 4, 1); /* Bank 1 OFF */
		mxl500x_set_gpio(state, 3, 1); /* Bank 2 OFF */
		mxl500x_set_gpio(state, 1, 0); /* Bank 3 ON  */
	}
	if (p->frequency >= 560000000 && p->frequency < 580000000) {
		MXL500x_SET_MAP(MXL500x_DAC2, DAC2_DAC_A_ENABLE, 0x1); /* Bank 4 ON */
		mxl500x_set_gpio(state, 4, 1); /* Bank 1 OFF */
		mxl500x_set_gpio(state, 3, 1); /* Bank 2 OFF */
		mxl500x_set_gpio(state, 1, 0); /* Bank 3 ON  */
	}
	if (p->frequency >= 580000000 && p->frequency < 630000000) {
		MXL500x_SET_MAP(MXL500x_DAC2, DAC2_DAC_A_ENABLE, 0x1); /* Bank 4 ON */
		mxl500x_set_gpio(state, 4, 1); /* Bank 1 OFF */
		mxl500x_set_gpio(state, 3, 1); /* Bank 2 OFF */
		mxl500x_set_gpio(state, 1, 0); /* Bank 3 ON  */
	}
	if (p->frequency >= 630000000 && p->frequency < 700000000) {
		MXL500x_SET_MAP(MXL500x_DAC2, DAC2_DAC_A_ENABLE, 0x1); /* Bank 4 ON */
		mxl500x_set_gpio(state, 4, 1); /* Bank 1 OFF */
		mxl500x_set_gpio(state, 3, 1); /* Bank 2 OFF */
		mxl500x_set_gpio(state, 1, 1); /* Bank 3 OFF */
	}
	if (p->frequency >= 700000000 && p->frequency < 760000000) {
		MXL500x_SET_MAP(MXL500x_DAC2, DAC2_DAC_A_ENABLE, 0x1); /* Bank 4 ON */
		mxl500x_set_gpio(state, 4, 1); /* Bank 1 OFF */
		mxl500x_set_gpio(state, 3, 1); /* Bank 2 OFF */
		mxl500x_set_gpio(state, 1, 1); /* Bank 3 OFF */
	}
	if (p->frequency >= 760000000 && p->frequency < 900000000) {
		MXL500x_SET_MAP(MXL500x_DAC2, DAC2_DAC_A_ENABLE, 0x1); /* Bank 4 ON */
		mxl500x_set_gpio(state, 4, 1); /* Bank 1 OFF */
		mxl500x_set_gpio(state, 3, 1); /* Bank 2 OFF */
		mxl500x_set_gpio(state, 1, 1); /* Bank 3 OFF */
	}
}

static void mxl500x_set_filterctl(struct mxl500x_state *state,
				  struct dvb_frontend_parameters *p)
{
	if (p->frequency < 43000000 || p->frequency > 900000000) {
		MXL500x_SET_MAP(MXL500x_DAC2, DAC2_DAC_A_ENABLE, 0x0);
		MXL500x_SET_MAP(MXL500x_DAC2, DAC2_DAC_B_ENABLE, 0x0);
	} else {
		MXL500x_SET_MAP(MXL500x_DAC2, DAC2_DAC_A_ENABLE, 0x1);
		MXL500x_SET_MAP(MXL500x_DAC2, DAC2_DAC_B_ENABLE, 0x1);
	}

	if (p->frequency >= 43000000 && p->frequency < 150000000) {
		MXL500x_SET_MAP(MXL500x_DAC1, DAC1_DAC_DIN_A, 63);
		MXL500x_SET_MAP(MXL500x_DAC2, DAC2_DAC_DIN_B, 0);
		mxl500x_set_gpio(state, 3, 0);
		mxl500x_set_gpio(state, 1, 1);
		mxl500x_set_gpio(state, 4, 1);
	}
	if (p->frequency >= 150000000 && p->frequency < 280000000) {
		MXL500x_SET_MAP(MXL500x_DAC1, DAC1_DAC_DIN_A, 63);
		MXL500x_SET_MAP(MXL500x_DAC2, DAC2_DAC_DIN_B, 0);
		mxl500x_set_gpio(state, 3, 1);
		mxl500x_set_gpio(state, 1, 0);
		mxl500x_set_gpio(state, 4, 1);
	}
	if (p->frequency >= 280000000 && p->frequency < 360000000) {
		MXL500x_SET_MAP(MXL500x_DAC1, DAC1_DAC_DIN_A, 63);
		MXL500x_SET_MAP(MXL500x_DAC2, DAC2_DAC_DIN_B, 0);
		mxl500x_set_gpio(state, 3, 1);
		mxl500x_set_gpio(state, 1, 0);
		mxl500x_set_gpio(state, 4, 0);
	}
	if (p->frequency >= 360000000 && p->frequency < 560000000) {
		MXL500x_SET_MAP(MXL500x_DAC1, DAC1_DAC_DIN_A, 63);
		MXL500x_SET_MAP(MXL500x_DAC2, DAC2_DAC_DIN_B, 0);
		mxl500x_set_gpio(state, 3, 1);
		mxl500x_set_gpio(state, 1, 1);
		mxl500x_set_gpio(state, 4, 0);
	}
	if (p->frequency >= 560000000 && p->frequency < 580000000) {
		MXL500x_SET_MAP(MXL500x_DAC1, DAC1_DAC_DIN_A, 0);
		MXL500x_SET_MAP(MXL500x_DAC2, DAC2_DAC_DIN_B, 29);
		mxl500x_set_gpio(state, 3, 1);
		mxl500x_set_gpio(state, 1, 1);
		mxl500x_set_gpio(state, 4, 0);
	}
	if (p->frequency >= 580000000 && p->frequency < 630000000) {
		MXL500x_SET_MAP(MXL500x_DAC1, DAC1_DAC_DIN_A, 0);
		MXL500x_SET_MAP(MXL500x_DAC2, DAC2_DAC_DIN_B, 0);
		mxl500x_set_gpio(state, 3, 1);
		mxl500x_set_gpio(state, 1, 1);
		mxl500x_set_gpio(state, 4, 0);
	}
	if (p->frequency >= 630000000 && p->frequency < 700000000) {
		MXL500x_SET_MAP(MXL500x_DAC1, DAC1_DAC_DIN_A, 0);
		MXL500x_SET_MAP(MXL500x_DAC2, DAC2_DAC_DIN_B, 16);
		mxl500x_set_gpio(state, 3, 1);
		mxl500x_set_gpio(state, 1, 1);
		mxl500x_set_gpio(state, 4, 1);
	}
	if (p->frequency >= 700000000 && p->frequency < 760000000) {
		MXL500x_SET_MAP(MXL500x_DAC1, DAC1_DAC_DIN_A, 0);
		MXL500x_SET_MAP(MXL500x_DAC2, DAC2_DAC_DIN_B, 7);
		mxl500x_set_gpio(state, 3, 1);
		mxl500x_set_gpio(state, 1, 1);
		mxl500x_set_gpio(state, 4, 1);
	}
	if (p->frequency >= 760000000 && p->frequency <= 900000000) {
		MXL500x_SET_MAP(MXL500x_DAC1, DAC1_DAC_DIN_A, 0);
		MXL500x_SET_MAP(MXL500x_DAC2, DAC2_DAC_DIN_B, 0);
		mxl500x_set_gpio(state, 3, 1);
		mxl500x_set_gpio(state, 1, 1);
		mxl500x_set_gpio(state, 4, 1);
	}
}

static u32 mxl_ceil(u32 val, u32 resol)
{
	return (val / resol + (val % resol > 0 ? 1:0));
}

static int mxl500x_set_params(struct dvb_frontend *fe,
			      struct dvb_frontend_parameters *p)
{
	struct mxl500x_state *state = fe->tuner_priv;
	const struct mxl500x_config *config = state->config;

	u32 int_mod, frac_mod, f_min, f_max, fmin_bin, fmax_bin, div;
	u32 eq_3, eq_4, eq_5, eq_5a, tg_lo, tg_div, tg, if_div;

	enum mxl500x_delsys delsys = MXL500x_MODE_DVBT; /* default */

#define MXL500x_XTAL_INT	(config->xtal_freq / 10000)

	switch (config->delsys) {
	case MXL500x_MODE_DVBT:
		break;
	case MXL500x_MODE_ATSC:
		switch (p->u.vsb.modulation) {
		case VSB_8:
		case VSB_16:
			delsys = MXL500x_MODE_ATSC;
			break;
		case QAM_64:
		case QAM_256:
		case QAM_AUTO:
			delsys = MXL500x_MODE_QAM;
			break;
		default:
			return -EINVAL;
		}
		break;
	default:
		return -EINVAL;
	}

	/* Use register defaults */
	memcpy(state->reg_field, reg_init, sizeof (reg_init));

	/* Step 1: Synthesizer RESET (Single AGC Mode) */
	printk("%s: Synthesizer RESET and latch\n", __func__);
	if (mxl500x_write(state, 0x09, 0xb1, MXL_LATCH))  /* master reg, synth reset */
		goto exit;

	/* Step 2: Block Init */
	// OVERRIDE_1, 2, 3, 4 Set to 1
	MXL500x_SET_MAP(MXL500x_MISC1, MISC1_OVERRIDE1, 1); /* Override 1 */
	MXL500x_SET_MAP(MXL500x_RFSYN4, RFSYN4_OVERRIDE2, 1); /* Override 2 */
	MXL500x_SET_MAP(MXL500x_RFSYN6, RFSYN6_OVERRIDE3, 1); /* Override 3 */
	MXL500x_SET_MAP(MXL500x_RFSYN7, RFSYN7_OVERRIDE4, 1); /* Override 4 */
	// Downconverter control
	MXL500x_SET_MAP(MXL500x_DC4, DC4_DN_IQTN_AMP_CUT, 1); /* Digital mode, Analog:0 */
	// Filter control
	MXL500x_SET_MAP(MXL500x_LPF1, LPF1_BB_MODE, 0); /* Digital mode, Analog:1 */
	MXL500x_SET_MAP(MXL500x_LPF1, LPF1_BB_BUF_0, 1); /* Digital mode, Analog: 0 */
	MXL500x_SET_MAP(MXL500x_LPF1, LPF1_BB_BUF_0A, 1); /* Digital mode, Analog: 0 */
	MXL500x_SET_MAP(MXL500x_LPF2, LPF2_BB_IQSWAP, 0); /* Digital mode, Analog: 1 */
	MXL500x_SET_MAP(MXL500x_LPF2, LPF2_BB_INITSTATE_DLPF_TUNE, 0);
	// Initialize LPF (Digital Mode)
	if (delsys == MXL500x_MODE_DVBT) {
		switch (p->u.ofdm.bandwidth) {
		case BANDWIDTH_8_MHZ:
			MXL500x_SET_MAP(MXL500x_LPF1, LPF1_BB_DLPF_BANDSEL, 0);
			break;
		case BANDWIDTH_7_MHZ:
			MXL500x_SET_MAP(MXL500x_LPF1, LPF1_BB_DLPF_BANDSEL, 2);
			break;
		case BANDWIDTH_6_MHZ:
			MXL500x_SET_MAP(MXL500x_LPF1, LPF1_BB_DLPF_BANDSEL, 3);
			break;
		default:
			printk("%s: Invalid Bandwidth mode specified %d\n", __func__, p->u.ofdm.bandwidth);
			return -EINVAL;
		}
	} else {
		/* ATSC/QAM = 6MHz default */
		MXL500x_SET_MAP(MXL500x_LPF1, LPF1_BB_DLPF_BANDSEL, 3);
	}
	// Charge pump control
	MXL500x_SET_MAP(MXL500x_RFSYN8, RFSYN8_RFSYN_CHP_GAIN, 5); /* Digital mode, Analog: 12 */
	MXL500x_SET_MAP(MXL500x_RFSYN8, RFSYN8_RFSYN_EN_CHP_HIGAIN, 1); /* Digital mode, Analog: 0 */
	// AGC TOP control
	MXL500x_SET_MAP(MXL500x_AGC1, AGC1_AGC_IF, 0xb); /* TOP=252 */
	MXL500x_SET_MAP(MXL500x_AGC1, AGC1_AGC_RF, 0xf);
	// DAC Tracking filter
	// IF Synthesizer control (Digital mode, 4.57MHz)
	MXL500x_SET_MAP(MXL500x_IFSYN4, IFSYN4_IF_DIVVAL, 0x06);
	MXL500x_SET_MAP(MXL500x_IFSYN5, IFSYN5_IF_VCO_BIAS, 0x08);
	int_mod = config->ref_freq / config->xtal_freq;
	MXL500x_SET_MAP(MXL500x_IFSYN1, IFSYN1_CHCAL_INT_MOD_IF, int_mod);
	frac_mod = (config->ref_freq / 1000) - (config->xtal_freq / 1000) * int_mod;
	frac_mod *= (2 << 15);
	frac_mod /= (config->xtal_freq / 1000);
	MXL500x_SET_MAP(MXL500x_IFSYN2, IFSYN2_CHCAL_FRAC_MOD_IF_MSB, (frac_mod >> 8));
	MXL500x_SET_MAP(MXL500x_IFSYN3, IFSYN3_CHCAL_FRAC_MOD_IF_LSB, frac_mod & 0xff);
	// IF Up converter control
	MXL500x_SET_MAP(MXL500x_IFUP1, IFUP1_DRV_RES_SEL, 6);
	MXL500x_SET_MAP(MXL500x_IFUP1, IFUP1_I_DRIVER, 2); // Load = 200 ohms
	// Anti alias filter control
	MXL500x_SET_MAP(MXL500x_IFUP1, IFUP1_EN_AAF, 1);
	MXL500x_SET_MAP(MXL500x_IFUP1, IFUP1_EN_3P, 1);
	MXL500x_SET_MAP(MXL500x_IFUP2, IFUP2_EN_AUX_3P, 1);
	MXL500x_SET_MAP(MXL500x_IFUP1, IFUP1_SEL_AAF_BAND, 0);
	// Demodulator clock output
	MXL500x_SET_MAP(MXL500x_DEMCLK, DEMCLK_SEQ_ENCLK16_CLK_OUT, 0); /* disable clock output */
	MXL500x_SET_MAP(MXL500x_DEMCLK, DEMCLK_SEQ_SEL4_16B, 1); /* output divider */
	// crystal control
	MXL500x_SET_MAP(MXL500x_XTAL1, XTAL1_XTAL_CAPSELECT, 1); /* cap enabled */
	MXL500x_SET_MAP(MXL500x_IFSYN4, IFSYN4_IF_SEL_DBL, 1); /* Xtal=16MHz */
	MXL500x_SET_MAP(MXL500x_RFSYN8, RFSYN8_RFSYN_R_DIV, 3); /* Xtal=16MHz */

	if (config->rssi_ena) {
		// RSSI control
		MXL500x_SET_MAP(MXL500x_RSSI4, RSSI4_SEQ_EXTSYNTHCALIF, 1);
		MXL500x_SET_MAP(MXL500x_DEMCLK, DEMCLK_SEQ_EXTDCCAL, 1);
		MXL500x_SET_MAP(MXL500x_RSSI5, RSSI5_AGC_EN_RSSI, 1);
		MXL500x_SET_MAP(MXL500x_RSSI1, RSSI1_RFA_ENCLKRFAGC, 1);
		// RSSI reference point
		MXL500x_SET_MAP(MXL500x_RSSI1, RSSI1_RFA_RSSI_REF, 2);
		MXL500x_SET_MAP(MXL500x_RSSI1, RSSI1_RFA_RSSI_REFH, 3);
		MXL500x_SET_MAP(MXL500x_RSSI2, RSSI2_RFA_RSSI_REFL, 1);
		// TOP
		MXL500x_SET_MAP(MXL500x_RSSI3, RSSI3_RFA_FLR, 0);
		MXL500x_SET_MAP(MXL500x_RSSI3, RSSI3_RFA_CEIL, 12);
	} else {
		// RSSI control
		MXL500x_SET_MAP(MXL500x_RSSI4, RSSI4_SEQ_EXTSYNTHCALIF, 0);
		MXL500x_SET_MAP(MXL500x_DEMCLK, DEMCLK_SEQ_EXTDCCAL, 0);
		MXL500x_SET_MAP(MXL500x_RSSI5, RSSI5_AGC_EN_RSSI, 0);
		MXL500x_SET_MAP(MXL500x_RSSI1, RSSI1_RFA_ENCLKRFAGC, 0);
	}

	// miscellaneous controls
	MXL500x_SET_MAP(MXL500x_MISC2, MISC2_SEQ_EXTIQFSMPULSE, 1); /* not analog low IF */

	MXL500x_SET_MAP(MXL500x_RSSI4, RSSI4_SEQ_EXTSYNTHCALIF, 1);
	MXL500x_SET_MAP(MXL500x_DEMCLK, DEMCLK_SEQ_EXTDCCAL, 1);
	MXL500x_SET_MAP(MXL500x_RSSI5, RSSI5_AGC_EN_RSSI, 1);
	MXL500x_SET_MAP(MXL500x_RSSI1, RSSI1_RFA_ENCLKRFAGC, 1);

	switch (delsys) {
	case MXL500x_MODE_DVBT: /* DVB-T Mode */
		/* RSSI reference point */
		MXL500x_SET_MAP(MXL500x_RSSI1, RSSI1_RFA_RSSI_REF, 3);
		MXL500x_SET_MAP(MXL500x_RSSI1, RSSI1_RFA_RSSI_REFH, 5);
		MXL500x_SET_MAP(MXL500x_RSSI2, RSSI2_RFA_RSSI_REFL, 1);
		/* TOP */
		MXL500x_SET_MAP(MXL500x_RSSI3, RSSI3_RFA_FLR, 2);
		MXL500x_SET_MAP(MXL500x_RSSI3, RSSI3_RFA_CEIL, 13);
		break;
	case MXL500x_MODE_ATSC: /* ATSC Mode */
		/* RSSI reference point */
		MXL500x_SET_MAP(MXL500x_RSSI1, RSSI1_RFA_RSSI_REF, 2);
		MXL500x_SET_MAP(MXL500x_RSSI1, RSSI1_RFA_RSSI_REFH, 4);
		MXL500x_SET_MAP(MXL500x_RSSI2, RSSI2_RFA_RSSI_REFL, 1);
		/* TOP */
		MXL500x_SET_MAP(MXL500x_RSSI3, RSSI3_RFA_FLR, 2);
		MXL500x_SET_MAP(MXL500x_RSSI3, RSSI3_RFA_CEIL, 13);
		break;
	case MXL500x_MODE_QAM: /* QAM Mode */
		/* RSSI reference point */
		MXL500x_SET_MAP(MXL500x_RSSI1, RSSI1_RFA_RSSI_REF, 3);
		MXL500x_SET_MAP(MXL500x_RSSI1, RSSI1_RFA_RSSI_REFH, 5);
		MXL500x_SET_MAP(MXL500x_RSSI2, RSSI2_RFA_RSSI_REFL, 2);

		MXL500x_SET_MAP(MXL500x_LPF1, LPF1_BB_MODE, 1);
		MXL500x_SET_MAP(MXL500x_LPF1, LPF1_BB_BUF_0, 2);
		MXL500x_SET_MAP(MXL500x_LPF1, LPF1_BB_BUF_0A, 0);

		MXL500x_SET_MAP(MXL500x_DC4, DC4_DN_IQTN_AMP_CUT, 0);
		MXL500x_SET_MAP(MXL500x_LPF1, LPF1_BB_ALPF_BANDSEL, 5);
		MXL500x_SET_MAP(MXL500x_RFSYN8, RFSYN8_RFSYN_CHP_GAIN, 2);
		break;
	default:
		return -EINVAL;
	}
	if (config->iflo_freq <= MXL500x_LOWIF_TH)
		MXL500x_SET_MAP(MXL500x_LPF2, LPF2_BB_IQSWAP, 0); /* Low IF */
	else
		MXL500x_SET_MAP(MXL500x_LPF2, LPF2_BB_IQSWAP, 1); /* High IF */

	//11, 12, 13, 22, 43, 44, 53, 56, 59, 73, 76, 77, 91, 134, 135, 137, 147, 156, 166, 167, 168
	// TODO! write registers (Init regs)
	printk("%s: Writing Init Regs\n", __func__);
	if (mxl500x_write_regs(state, mxl500x_init_regs, sizeof(mxl500x_init_regs)))
		goto exit;

	/* Step 3: ZIF Mode */
	// Synthesizer reset
	printk("%s: Synthesizer RESET and latch\n", __func__);
	if (mxl500x_write(state, 0x09, 0xb1, MXL_LATCH))  /* master reg, synth reset, latch */
		goto exit;

	if (p->frequency < 40000000)
		return -EINVAL;

	if (p->frequency >= 40000000 && p->frequency <= 75000000) {
		MXL500x_SET_MAP(MXL500x_DC1, DC1_DN_POLY, 2);
		MXL500x_SET_MAP(MXL500x_DC3, DC3_DN_RFGAIN, 3);
		MXL500x_SET_MAP(MXL500x_DC2, DC2_DN_CAP_RFLPF_0, (MXL500x_MSB(DC2_DN_CAP_RFLPF_0, 423)));
		MXL500x_SET_MAP(MXL500x_DC1, DC1_DN_CAP_RFLPF_1, (MXL500x_LSB(DC2_DN_CAP_RFLPF_0, 423)));
		MXL500x_SET_MAP(MXL500x_DC3, DC3_DN_EN_VHFUHFBAR, 1);
		MXL500x_SET_MAP(MXL500x_DC4, DC4_DN_GAIN_ADJUST, 1);
	}
	if (p->frequency > 75000000 && p->frequency <= 100000000) {
		MXL500x_SET_MAP(MXL500x_DC1, DC1_DN_POLY, 3);
		MXL500x_SET_MAP(MXL500x_DC3, DC3_DN_RFGAIN, 3);
		MXL500x_SET_MAP(MXL500x_DC2, DC2_DN_CAP_RFLPF_0, (MXL500x_MSB(DC2_DN_CAP_RFLPF_0, 222)));
		MXL500x_SET_MAP(MXL500x_DC1, DC1_DN_CAP_RFLPF_1, (MXL500x_LSB(DC2_DN_CAP_RFLPF_0, 222)));
		MXL500x_SET_MAP(MXL500x_DC3, DC3_DN_EN_VHFUHFBAR, 1);
		MXL500x_SET_MAP(MXL500x_DC4, DC4_DN_GAIN_ADJUST, 1);
	}
	if (p->frequency > 100000000 && p->frequency <= 150000000) {
		MXL500x_SET_MAP(MXL500x_DC1, DC1_DN_POLY, 3);
		MXL500x_SET_MAP(MXL500x_DC3, DC3_DN_RFGAIN, 3);
		MXL500x_SET_MAP(MXL500x_DC2, DC2_DN_CAP_RFLPF_0, (MXL500x_MSB(DC2_DN_CAP_RFLPF_0, 147)));
		MXL500x_SET_MAP(MXL500x_DC1, DC1_DN_CAP_RFLPF_1, (MXL500x_LSB(DC2_DN_CAP_RFLPF_0, 147)));
		MXL500x_SET_MAP(MXL500x_DC3, DC3_DN_EN_VHFUHFBAR, 1);
		MXL500x_SET_MAP(MXL500x_DC4, DC4_DN_GAIN_ADJUST, 2);
	}
	if (p->frequency > 150000000 && p->frequency <= 200000000) {
		MXL500x_SET_MAP(MXL500x_DC1, DC1_DN_POLY, 3);
		MXL500x_SET_MAP(MXL500x_DC3, DC3_DN_RFGAIN, 3);
		MXL500x_SET_MAP(MXL500x_DC2, DC2_DN_CAP_RFLPF_0, (MXL500x_MSB(DC2_DN_CAP_RFLPF_0, 9)));
		MXL500x_SET_MAP(MXL500x_DC1, DC1_DN_CAP_RFLPF_1, (MXL500x_LSB(DC2_DN_CAP_RFLPF_0, 9)));
		MXL500x_SET_MAP(MXL500x_DC3, DC3_DN_EN_VHFUHFBAR, 1);
		MXL500x_SET_MAP(MXL500x_DC4, DC4_DN_GAIN_ADJUST, 2);
	}
	if (p->frequency > 200000000 && p->frequency <= 300000000) {
		MXL500x_SET_MAP(MXL500x_DC1, DC1_DN_POLY, 3);
		MXL500x_SET_MAP(MXL500x_DC3, DC3_DN_RFGAIN, 3);
		MXL500x_SET_MAP(MXL500x_DC2, DC2_DN_CAP_RFLPF_0, (MXL500x_MSB(DC2_DN_CAP_RFLPF_0, 0)));
		MXL500x_SET_MAP(MXL500x_DC1, DC1_DN_CAP_RFLPF_1, (MXL500x_LSB(DC2_DN_CAP_RFLPF_0, 0)));
		MXL500x_SET_MAP(MXL500x_DC3, DC3_DN_EN_VHFUHFBAR, 1);
		MXL500x_SET_MAP(MXL500x_DC4, DC4_DN_GAIN_ADJUST, 3);
	}
	if (p->frequency > 300000000 && p->frequency <= 650000000) {
		MXL500x_SET_MAP(MXL500x_DC1, DC1_DN_POLY, 3);
		MXL500x_SET_MAP(MXL500x_DC3, DC3_DN_RFGAIN, 1);
		MXL500x_SET_MAP(MXL500x_DC2, DC2_DN_CAP_RFLPF_0, (MXL500x_MSB(DC2_DN_CAP_RFLPF_0, 0)));
		MXL500x_SET_MAP(MXL500x_DC1, DC1_DN_CAP_RFLPF_1, (MXL500x_LSB(DC2_DN_CAP_RFLPF_0, 0)));
		MXL500x_SET_MAP(MXL500x_DC3, DC3_DN_EN_VHFUHFBAR, 0);
		MXL500x_SET_MAP(MXL500x_DC4, DC4_DN_GAIN_ADJUST, 3);
	}
	if (p->frequency > 650000000 && p->frequency <= 900000000) {
		MXL500x_SET_MAP(MXL500x_DC1, DC1_DN_POLY, 3);
		MXL500x_SET_MAP(MXL500x_DC3, DC3_DN_RFGAIN, 2);
		MXL500x_SET_MAP(MXL500x_DC2, DC2_DN_CAP_RFLPF_0, (MXL500x_MSB(DC2_DN_CAP_RFLPF_0, 0)));
		MXL500x_SET_MAP(MXL500x_DC1, DC1_DN_CAP_RFLPF_1, (MXL500x_LSB(DC2_DN_CAP_RFLPF_0, 0)));
		MXL500x_SET_MAP(MXL500x_DC3, DC3_DN_EN_VHFUHFBAR, 0);
		MXL500x_SET_MAP(MXL500x_DC4, DC4_DN_GAIN_ADJUST, 3);
	}

	if (p->frequency > 40000000 && p->frequency <= 800000000) {
		MXL500x_SET_MAP(MXL500x_DC3, DC3_DN_IQTNBUF_AMP, 1);
		MXL500x_SET_MAP(MXL500x_DC3, DC3_DN_IQTNGBFBIAS_BST, 0);
	}
	if (p->frequency > 800000000 && p->frequency <= 900000000) {
		MXL500x_SET_MAP(MXL500x_DC3, DC3_DN_IQTNBUF_AMP, 10);
		MXL500x_SET_MAP(MXL500x_DC3, DC3_DN_IQTNGBFBIAS_BST, 1);
	}
	// Set RF synthesizer and LO path control
	fmin_bin = 28000000, fmax_bin = 42500000;
	if (p->frequency >= 40000000 && p->frequency <= fmax_bin) {
		MXL500x_SET_MAP(MXL500x_RFSYN6, RFSYN6_RFSYN_EN_OUTMUX, 1);
		MXL500x_SET_MAP(MXL500x_RFSYN6, RFSYN6_RFSYN_SEL_VCO_OUT, 0);
		MXL500x_SET_MAP(MXL500x_RFSYN6, RFSYN6_RFSYN_SEL_VCO_HI, 0);
		MXL500x_SET_MAP(MXL500x_RFSYN6, RFSYN6_RFSYN_SEL_DIVM, 0);
		MXL500x_SET_MAP(MXL500x_RFSYN5, RFSYN5_RFSYN_RF_DIV_BIAS, 1);
		MXL500x_SET_MAP(MXL500x_DC2, DC2_DN_SEL_FREQ, 1);
		f_min = fmin_bin, f_max = fmax_bin, div = 64;
	}
	fmin_bin = 42500000, fmax_bin = 56000000;
	if (p->frequency >= fmin_bin && p->frequency <= fmax_bin) {
		MXL500x_SET_MAP(MXL500x_RFSYN6, RFSYN6_RFSYN_EN_OUTMUX, 1);
		MXL500x_SET_MAP(MXL500x_RFSYN6, RFSYN6_RFSYN_SEL_VCO_OUT, 0);
		MXL500x_SET_MAP(MXL500x_RFSYN6, RFSYN6_RFSYN_SEL_VCO_HI, 1);
		MXL500x_SET_MAP(MXL500x_RFSYN6, RFSYN6_RFSYN_SEL_DIVM, 0);
		MXL500x_SET_MAP(MXL500x_RFSYN5, RFSYN5_RFSYN_RF_DIV_BIAS, 1);
		MXL500x_SET_MAP(MXL500x_DC2, DC2_DN_SEL_FREQ, 1);
		f_min = fmin_bin, f_max = fmax_bin, div = 64;
	}
	fmin_bin = 56000000, fmax_bin = 85000000;
	if (p->frequency >= fmin_bin && p->frequency <= fmax_bin) {
		MXL500x_SET_MAP(MXL500x_RFSYN6, RFSYN6_RFSYN_EN_OUTMUX, 0);
		MXL500x_SET_MAP(MXL500x_RFSYN6, RFSYN6_RFSYN_SEL_VCO_OUT, 1);
		MXL500x_SET_MAP(MXL500x_RFSYN6, RFSYN6_RFSYN_SEL_VCO_HI, 0);
		MXL500x_SET_MAP(MXL500x_RFSYN6, RFSYN6_RFSYN_SEL_DIVM, 0);
		MXL500x_SET_MAP(MXL500x_RFSYN5, RFSYN5_RFSYN_RF_DIV_BIAS, 1);
		MXL500x_SET_MAP(MXL500x_DC2, DC2_DN_SEL_FREQ, 1);
		f_min = fmin_bin, f_max = fmax_bin, div = 32;
	}
	fmin_bin = 85000000, fmax_bin = 112000000;
	if (p->frequency >= fmin_bin && p->frequency <= fmax_bin) {
		MXL500x_SET_MAP(MXL500x_RFSYN6, RFSYN6_RFSYN_EN_OUTMUX, 0);
		MXL500x_SET_MAP(MXL500x_RFSYN6, RFSYN6_RFSYN_SEL_VCO_OUT, 1);
		MXL500x_SET_MAP(MXL500x_RFSYN6, RFSYN6_RFSYN_SEL_VCO_HI, 1);
		MXL500x_SET_MAP(MXL500x_RFSYN6, RFSYN6_RFSYN_SEL_DIVM, 0);
		MXL500x_SET_MAP(MXL500x_RFSYN5, RFSYN5_RFSYN_RF_DIV_BIAS, 1);
		MXL500x_SET_MAP(MXL500x_DC2, DC2_DN_SEL_FREQ, 1);
		f_min = fmin_bin, f_max = fmax_bin, div = 32;
	}
	fmin_bin = 112000000, fmax_bin = 170000000;
	if (p->frequency >= fmin_bin && p->frequency <= fmax_bin) {
		MXL500x_SET_MAP(MXL500x_RFSYN6, RFSYN6_RFSYN_EN_OUTMUX, 0);
		MXL500x_SET_MAP(MXL500x_RFSYN6, RFSYN6_RFSYN_SEL_VCO_OUT, 1);
		MXL500x_SET_MAP(MXL500x_RFSYN6, RFSYN6_RFSYN_SEL_VCO_HI, 0);
		MXL500x_SET_MAP(MXL500x_RFSYN6, RFSYN6_RFSYN_SEL_DIVM, 0);
		MXL500x_SET_MAP(MXL500x_RFSYN5, RFSYN5_RFSYN_RF_DIV_BIAS, 1);
		MXL500x_SET_MAP(MXL500x_DC2, DC2_DN_SEL_FREQ, 2);
		f_min = fmin_bin, f_max = fmax_bin, div = 16;
	}
	fmin_bin = 170000000, fmax_bin = 225000000;
	if (p->frequency >= fmin_bin && p->frequency <= fmax_bin) {
		MXL500x_SET_MAP(MXL500x_RFSYN6, RFSYN6_RFSYN_EN_OUTMUX, 0);
		MXL500x_SET_MAP(MXL500x_RFSYN6, RFSYN6_RFSYN_SEL_VCO_OUT, 1);
		MXL500x_SET_MAP(MXL500x_RFSYN6, RFSYN6_RFSYN_SEL_VCO_HI, 1);
		MXL500x_SET_MAP(MXL500x_RFSYN6, RFSYN6_RFSYN_SEL_DIVM, 0);
		MXL500x_SET_MAP(MXL500x_RFSYN5, RFSYN5_RFSYN_RF_DIV_BIAS, 1);
		MXL500x_SET_MAP(MXL500x_DC2, DC2_DN_SEL_FREQ, 2);
		f_min = fmin_bin, f_max = fmax_bin, div = 16;
	}
	fmin_bin = 225000000, fmax_bin = 300000000;
	if (p->frequency >= fmin_bin && p->frequency <= fmax_bin) {
		MXL500x_SET_MAP(MXL500x_RFSYN6, RFSYN6_RFSYN_EN_OUTMUX, 0);
		MXL500x_SET_MAP(MXL500x_RFSYN6, RFSYN6_RFSYN_SEL_VCO_OUT, 1);
		MXL500x_SET_MAP(MXL500x_RFSYN6, RFSYN6_RFSYN_SEL_VCO_HI, 0);
		MXL500x_SET_MAP(MXL500x_RFSYN6, RFSYN6_RFSYN_SEL_DIVM, 0);
		MXL500x_SET_MAP(MXL500x_RFSYN5, RFSYN5_RFSYN_RF_DIV_BIAS, 1);
		MXL500x_SET_MAP(MXL500x_DC2, DC2_DN_SEL_FREQ, 4);
		f_min = fmin_bin, f_max = fmax_bin, div = 8;
	}
	fmin_bin = 225000000, fmax_bin = 300000000;
	if (p->frequency >= fmin_bin && p->frequency <= fmax_bin) {
		MXL500x_SET_MAP(MXL500x_RFSYN6, RFSYN6_RFSYN_EN_OUTMUX, 0);
		MXL500x_SET_MAP(MXL500x_RFSYN6, RFSYN6_RFSYN_SEL_VCO_OUT, 1);
		MXL500x_SET_MAP(MXL500x_RFSYN6, RFSYN6_RFSYN_SEL_VCO_HI, 0);
		MXL500x_SET_MAP(MXL500x_RFSYN6, RFSYN6_RFSYN_SEL_DIVM, 0);
		MXL500x_SET_MAP(MXL500x_RFSYN5, RFSYN5_RFSYN_RF_DIV_BIAS, 1);
		MXL500x_SET_MAP(MXL500x_DC2, DC2_DN_SEL_FREQ, 4);
		f_min = fmin_bin, f_max = 340000000, div = 8;
	}
	fmin_bin = 300000000, fmax_bin = 340000000;
	if (p->frequency >= fmin_bin && p->frequency <= fmax_bin) {
		MXL500x_SET_MAP(MXL500x_RFSYN6, RFSYN6_RFSYN_EN_OUTMUX, 1);
		MXL500x_SET_MAP(MXL500x_RFSYN6, RFSYN6_RFSYN_SEL_VCO_OUT, 0);
		MXL500x_SET_MAP(MXL500x_RFSYN6, RFSYN6_RFSYN_SEL_VCO_HI, 0);
		MXL500x_SET_MAP(MXL500x_RFSYN6, RFSYN6_RFSYN_SEL_DIVM, 0);
		MXL500x_SET_MAP(MXL500x_RFSYN5, RFSYN5_RFSYN_RF_DIV_BIAS, 1);
		MXL500x_SET_MAP(MXL500x_DC2, DC2_DN_SEL_FREQ, 0);
		f_min = 225000000, f_max = fmax_bin, div = 8;
	}
	fmin_bin = 340000000, fmax_bin = 450000000;
	if (p->frequency >= fmin_bin && p->frequency <= fmax_bin) {
		MXL500x_SET_MAP(MXL500x_RFSYN6, RFSYN6_RFSYN_EN_OUTMUX, 1);
		MXL500x_SET_MAP(MXL500x_RFSYN6, RFSYN6_RFSYN_SEL_VCO_OUT, 0);
		MXL500x_SET_MAP(MXL500x_RFSYN6, RFSYN6_RFSYN_SEL_VCO_HI, 1);
		MXL500x_SET_MAP(MXL500x_RFSYN6, RFSYN6_RFSYN_SEL_DIVM, 0);
		MXL500x_SET_MAP(MXL500x_RFSYN5, RFSYN5_RFSYN_RF_DIV_BIAS, 2);
		MXL500x_SET_MAP(MXL500x_DC2, DC2_DN_SEL_FREQ, 0);
		f_min = fmin_bin, f_max = fmax_bin, div = 8;
	}
	fmin_bin = 450000000, fmax_bin = 680000000;
	if (p->frequency >= fmin_bin && p->frequency <= fmax_bin) {
		MXL500x_SET_MAP(MXL500x_RFSYN6, RFSYN6_RFSYN_EN_OUTMUX, 0);
		MXL500x_SET_MAP(MXL500x_RFSYN6, RFSYN6_RFSYN_SEL_VCO_OUT, 1);
		MXL500x_SET_MAP(MXL500x_RFSYN6, RFSYN6_RFSYN_SEL_VCO_HI, 0);
		MXL500x_SET_MAP(MXL500x_RFSYN6, RFSYN6_RFSYN_SEL_DIVM, 1);
		MXL500x_SET_MAP(MXL500x_RFSYN5, RFSYN5_RFSYN_RF_DIV_BIAS, 1);
		MXL500x_SET_MAP(MXL500x_DC2, DC2_DN_SEL_FREQ, 0);
		f_min = fmin_bin, f_max = fmax_bin, div = 4;
	}
	fmin_bin = 680000000, fmax_bin = 900000000;
	if (p->frequency >= fmin_bin && p->frequency <= fmax_bin) {
		MXL500x_SET_MAP(MXL500x_RFSYN6, RFSYN6_RFSYN_EN_OUTMUX, 0);
		MXL500x_SET_MAP(MXL500x_RFSYN6, RFSYN6_RFSYN_SEL_VCO_OUT, 1);
		MXL500x_SET_MAP(MXL500x_RFSYN6, RFSYN6_RFSYN_SEL_VCO_HI, 1);
		MXL500x_SET_MAP(MXL500x_RFSYN6, RFSYN6_RFSYN_SEL_DIVM, 1);
		MXL500x_SET_MAP(MXL500x_RFSYN5, RFSYN5_RFSYN_RF_DIV_BIAS, 1);
		MXL500x_SET_MAP(MXL500x_DC2, DC2_DN_SEL_FREQ, 0);
		f_min = fmin_bin, f_max = fmax_bin, div = 4;
	}
	// Equation 3 RF synthesizer VCO Bias
	eq_3 = (((f_max - p->frequency) / 1000) * 32) / ((f_max - f_min) / 1000) + 8;
	MXL500x_SET_MAP(MXL500x_RFSYN5, RFSYN5_RFSYN_VCO_BIAS, eq_3);
	// Equation 4
	eq_4 = (p->frequency * div / 1000) / (2 * config->xtal_freq * 2 / 1000);
	MXL500x_SET_MAP(MXL500x_RFSYN1, RFSYN1_CHCAL_INT_MOD_RF, eq_4);
	// Equation 5
	eq_5 = ((2 << 17) * (p->frequency / 10000 * div - (eq_4 * (2 * config->xtal_freq * 2) / 10000))) / (2 * config->xtal_freq * 2 / 10000);

	MXL500x_SET_MAP(MXL500x_RFSYN2, RFSYN2_CHCAL_FRAC_MOD_RF_2, ((eq_5 >> 10) & 0xff));
	MXL500x_SET_MAP(MXL500x_RFSYN3, RFSYN3_CHCAL_FRAC_MOD_RF_1, ((eq_5 >>  2) & 0xff));
	MXL500x_SET_MAP(MXL500x_RFSYN4, RFSYN4_CHCAL_FRAC_MOD_RF_0, (eq_5 & 0x03));
	// Equation E5A
	eq_5a = (((f_max - p->frequency) / 1000) * 4 / ((f_max - f_min) / 1000)) + 1;
	MXL500x_SET_MAP(MXL500x_RFSYN7, RFSYN7_RFSYN_LPF_R, eq_5a);
	MXL500x_SET_MAP(MXL500x_RFSYN1, RFSYN1_CHCAL_EN_INT_RF, ((eq_5 == 0) ? 1:0));
	// Set TG synthesizer
	tg_lo = p->frequency - 750000;
	if (tg_lo < 33000000)
		return -1;

	fmin_bin = 33000000, fmax_bin = 50000000;
	if (tg_lo >= fmin_bin && tg_lo <= fmax_bin) {
		MXL500x_SET_MAP(MXL500x_IQCALSYN2, IQCALSYN2_TG_LO_DIVVAL, 0x6);
		MXL500x_SET_MAP(MXL500x_IQCALSYN1, IQCALSYN1_TG_LO_SELVAL_1, 0x0);
		MXL500x_SET_MAP(MXL500x_IQCALSYN2, IQCALSYN2_TG_LO_SELVAL_0, 0x0);
		f_min = fmin_bin, f_max = fmax_bin, div = 36;
	}
	fmin_bin = 50000000, fmax_bin = 67000000;
	if (tg_lo > fmin_bin && tg_lo <= fmax_bin) {
		MXL500x_SET_MAP(MXL500x_IQCALSYN2, IQCALSYN2_TG_LO_DIVVAL, 0x1);
		MXL500x_SET_MAP(MXL500x_IQCALSYN1, IQCALSYN1_TG_LO_SELVAL_1, 0x0);
		MXL500x_SET_MAP(MXL500x_IQCALSYN2, IQCALSYN2_TG_LO_SELVAL_0, 0x0);
		f_min = fmin_bin, f_max = fmax_bin, div = 24;
	}
	fmin_bin = 67000000, fmax_bin = 100000000;
	if (tg_lo > fmin_bin && tg_lo <= fmax_bin) {
		MXL500x_SET_MAP(MXL500x_IQCALSYN2, IQCALSYN2_TG_LO_DIVVAL, 0xc);
		MXL500x_SET_MAP(MXL500x_IQCALSYN1, IQCALSYN1_TG_LO_SELVAL_1,0x1);
		MXL500x_SET_MAP(MXL500x_IQCALSYN2, IQCALSYN2_TG_LO_SELVAL_0, 0x0);
		f_min = fmin_bin, f_max = fmax_bin, div = 18;
	}
	fmin_bin = 100000000, fmax_bin = 150000000;
	if (tg_lo > fmin_bin && tg_lo <= fmax_bin) {
		MXL500x_SET_MAP(MXL500x_IQCALSYN2, IQCALSYN2_TG_LO_DIVVAL, 0x8);
		MXL500x_SET_MAP(MXL500x_IQCALSYN1, IQCALSYN1_TG_LO_SELVAL_1,0x1);
		MXL500x_SET_MAP(MXL500x_IQCALSYN2, IQCALSYN2_TG_LO_SELVAL_0, 0x0);
		f_min = fmin_bin, f_max = fmax_bin, div = 12;
	}
	fmin_bin = 150000000, fmax_bin = 200000000;
	if (tg_lo > fmin_bin && tg_lo <= fmax_bin) {
		MXL500x_SET_MAP(MXL500x_IQCALSYN2, IQCALSYN2_TG_LO_DIVVAL, 0x0);
		MXL500x_SET_MAP(MXL500x_IQCALSYN1, IQCALSYN1_TG_LO_SELVAL_1,0x1);
		MXL500x_SET_MAP(MXL500x_IQCALSYN2, IQCALSYN2_TG_LO_SELVAL_0, 0x0);
		f_min = fmin_bin, f_max = fmax_bin, div = 8;
	}
	fmin_bin = 200000000, fmax_bin = 300000000;
	if (tg_lo > fmin_bin && tg_lo <= fmax_bin) {
		MXL500x_SET_MAP(MXL500x_IQCALSYN2, IQCALSYN2_TG_LO_DIVVAL, 0x8);
		MXL500x_SET_MAP(MXL500x_IQCALSYN1, IQCALSYN1_TG_LO_SELVAL_1,0x1);
		MXL500x_SET_MAP(MXL500x_IQCALSYN2, IQCALSYN2_TG_LO_SELVAL_0, 0x1);
		f_min = fmin_bin, f_max = fmax_bin, div = 6;
	}
	fmin_bin = 300000000, fmax_bin = 400000000;
	if (tg_lo > fmin_bin && tg_lo <= fmax_bin) {
		MXL500x_SET_MAP(MXL500x_IQCALSYN2, IQCALSYN2_TG_LO_DIVVAL, 0x0);
		MXL500x_SET_MAP(MXL500x_IQCALSYN1, IQCALSYN1_TG_LO_SELVAL_1,0x1);
		MXL500x_SET_MAP(MXL500x_IQCALSYN2, IQCALSYN2_TG_LO_SELVAL_0, 0x1);
		f_min = fmin_bin, f_max = fmax_bin, div = 4;
	}
	fmin_bin = 400000000, fmax_bin = 600000000;
	if (tg_lo > fmin_bin && tg_lo <= fmax_bin) {
		MXL500x_SET_MAP(MXL500x_IQCALSYN2, IQCALSYN2_TG_LO_DIVVAL, 0x8);
		MXL500x_SET_MAP(MXL500x_IQCALSYN1, IQCALSYN1_TG_LO_SELVAL_1,0x3);
		MXL500x_SET_MAP(MXL500x_IQCALSYN2, IQCALSYN2_TG_LO_SELVAL_0, 0x1);
		f_min = fmin_bin, f_max = fmax_bin, div = 3;
	}
	fmin_bin = 600000000, fmax_bin = 900000000;
	if (tg_lo > fmin_bin && tg_lo <= fmax_bin) {
		MXL500x_SET_MAP(MXL500x_IQCALSYN2, IQCALSYN2_TG_LO_DIVVAL, 0x0);
		MXL500x_SET_MAP(MXL500x_IQCALSYN1, IQCALSYN1_TG_LO_SELVAL_1,0x3);
		MXL500x_SET_MAP(MXL500x_IQCALSYN2, IQCALSYN2_TG_LO_SELVAL_0, 0x1);
		f_min = fmin_bin, f_max = fmax_bin, div = 2;
	}

	// TG_DIV_VAL
	tg_div = (tg_lo * div / 100000) * (mxl_ceil(config->xtal_freq, 1000000) * 100) / (config->xtal_freq / 1000);

	if (tg_lo > 600000000) {
		tg_div += 1;
	}
	MXL500x_SET_MAP(MXL500x_IQCALSYN3, IQCALSYN3_TG_DIV_VAL, (MXL500x_MSB(IQCALSYN4_TG_DIV_VAL, tg_div)));
	MXL500x_SET_MAP(MXL500x_IQCALSYN4, IQCALSYN4_TG_DIV_VAL, (MXL500x_LSB(IQCALSYN4_TG_DIV_VAL, tg_div)));

	f_max = 180000, f_min = 120000;
	tg = (f_max * MXL500x_XTAL_INT - (tg_lo / 10000) * div * MXL500x_XTAL_INT) * 32 / ((f_max - f_min) * MXL500x_XTAL_INT) + 8;
	MXL500x_SET_MAP(MXL500x_IQCALSYN1, IQCALSYN1_TG_VCO_BIAS, tg);
	switch (config->octf) {
	case MXL500x_OCTF_OFF:
		mxl500x_octfilter_off(state, p);
		break;
	case MXL500x_OCTF_C:
		mxl500x_octfilter_c(state, p);
		break;
	case MXL500x_OCTF_CH:
		mxl500x_octfilter_ch(state, p);
		break;
	default:
		mxl500x_set_filterctl(state, p);
		break;
	}

	/* Step 4 */
	// retrieve if divider
	if_div = MXL500x_GETFIELD(IFSYN4_IF_DIVVAL, mxl500x_get_reg(state, MXL500x_IFSYN4));
	MXL500x_SET_MAP(MXL500x_MISC_TUNE1, MISC_TUNE1_SEQ_FSM_PULSE, 0);
	MXL500x_SET_MAP(MXL500x_MISC_TUNE2, MISC_TUNE2_SEQ_EXTPOWERUP, 1);
	MXL500x_SET_MAP(MXL500x_IFSYN4, IFSYN4_IF_DIVVAL, 8);
	// Synthesizer LOAD Start
//	printk("%s: Synthesizer Load START\n", __func__);
//	if (mxl500x_write(state, 0x09, 0xf2, MXL_NO_LATCH)) /* master reg, load start, don't latch */
//		goto exit;
	// write all changed regs (change regs)
	// 14, 15, 16, 17, 22, 43, 68, 69, 70, 73, 92, 93, 106, 107, 108, 109, 110, 111, 112, 136, 138, 149
	mxl500x_set_reg(state, 0x09, 0xf3);
	printk("%s: Setup changed registers\n", __func__);
	if (mxl500x_write_regs(state, mxl500x_zif_regs, sizeof(mxl500x_zif_regs)))
		goto exit;

	msleep(50); /* wait for calibration */
	MXL500x_SET_MAP(MXL500x_MISC_TUNE1, MISC_TUNE1_SEQ_FSM_PULSE, 1);
	MXL500x_SET_MAP(MXL500x_IFSYN4, IFSYN4_IF_DIVVAL, if_div);
	// Synthesizer LOAD Start
//	printk("%s: Synthesizer Load START\n", __func__);
//	if (mxl500x_write(state, 0x09, 0xf2, MXL_NO_LATCH)) /* master reg, load start, don't latch */
//		goto exit;
	// write regs
	// 43, 136
	printk("%s: Tuner go\n", __func__);
	if (mxl500x_write_regs(state, mxl500x_go_regs, sizeof(mxl500x_go_regs)))
		goto exit;

	printk("%s: Done\n", __func__);

	return 0;

exit:
	printk("%s: I/O error\n", __func__);
	return -EIO;
}

static const struct dvb_tuner_ops mxl500x_ops = {
	.info = {
		.name		= "Maxlinear MXL500x",
		.frequency_min	=  44000000,
		.frequency_max	= 885000000,
	},

	.release	= mxl500x_release,
	.init		= mxl500x_init,

	.set_params	= mxl500x_set_params,
};

struct dvb_frontend *mxl500x_attach(struct dvb_frontend *fe,
				    const struct mxl500x_config *config,
				    struct i2c_adapter *i2c)
{
	struct mxl500x_state *state;

	printk("%s: Attaching ...\n", __func__);
	if ((state = kzalloc(sizeof (struct mxl500x_state), GFP_KERNEL)) == NULL) {
		fe = NULL;
		goto exit;
	}
	state->config	= config;
	state->frontend	= fe;
	state->i2c	= i2c;

	memcpy(&fe->ops.tuner_ops, &mxl500x_ops, sizeof (struct dvb_tuner_ops));
	fe->tuner_priv	= state;

	printk("%s: MXL500x tuner succesfully attached\n", __func__);
	return fe;

exit:
	printk("%s: Error attaching tuner\n", __func__);
	return NULL;
}
EXPORT_SYMBOL(mxl500x_attach);

MODULE_AUTHOR("Manu Abraham <abraham.manu@gmail.com>");
MODULE_AUTHOR("Matthias Schwarzott <zzam@gentoo.org>");
MODULE_DESCRIPTION("Maxlinear MXL5003 driver");
MODULE_LICENSE("GPL");
